//
// o r g a n . c
//
// A keyboard organ program
//
// Dec. 5, 2002 by W.Nishida (http://www.skyfree.org)
//

#include <stdio.h>		// printf()
#include <sys/io.h>		// ioperm(), inb(), outb()
#include <termios.h>		// termios{}, tcgetattr(), tcsetattr()
#include <unistd.h>		// STDIN_FILENO
#include <sys/ioctl.h>		// ioctl()
#include <linux/kd.h>		// KDSKBMODE, K_XLATE, K_MEDIUMRAW

struct termios tm, ntm;		// "termios" structure holders

#define LINUXSCAN_ESC	1	// Linux scan code of "ESC"
#define LINUXSCAN_A	30	// Linux scan code of "A"
#define LINUXSCAN_K	37	// Linux scan code of "K"

// Speaker control parameters

#define PORT_PIT_CTL	0x43	// 8253 Programmable Interval Timer control
#define PORT_PIT_CNT2	0x42	// PIT 2nd counter
#define PORT_8255B	0x61	// 8255 port B
#define PORT_WAIT	0x80	// 1 microsec wait / access
#define PIT_CMD		0xB6	// PIT control word
#define PIT_CLK2	1193180	// PIT clock frequency

// Sound frequency table (Hz)

int snd_tbl[] = {
  264,		// C4
  297,		// D4
  330,		// E4
  352,		// F4
  396,		// G4
  440,		// A4
  495,		// B4
  528		// C5
 };

// simple beep function
//	hz: sound frequency

void beep(int hz) {
  int cnt, last;
  static int current;

  // Zero turns off speaker

  if (hz == 0) {			// Be silent
    last = inb(PORT_8255B);		// Read last output value
    outb(last & ~3, PORT_8255B);	// Speaker off
    current = 0;			// Reset current frequency
    return;
   }
  else if (hz == current)		// Same frequency?
    return;				// Then, do nothing
  else
    current = hz;			// Update work

  // Convert sound frequency to timer2 count

  cnt = PIT_CLK2 / hz;

  // Define sound frequency

  outb(PIT_CMD, PORT_PIT_CTL);		// Send a command to PIT
  inb(PORT_WAIT);			// Wait
  outb(cnt & 0xff, PORT_PIT_CNT2);	// Set LSB of counter 2
  inb(PORT_WAIT);			// Wait
  outb((cnt&0xff00)>>8, PORT_PIT_CNT2);	// Set MSB of Counter 2

  // Speaker control

  last = inb(PORT_8255B);		// Read current value
  outb(last | 3, PORT_8255B);		// Speaker on
 }

int main(int argc, char** argv) {
  int ret, up = 0, scan, idx, bit;
  unsigned char key, stat = 0;

  // Allow I/O access against PIT, port B and 0x80

  ret  = ioperm(PORT_PIT_CTL, 1, 1);
  ret |= ioperm(PORT_PIT_CNT2, 1, 1);
  ret |= ioperm(PORT_8255B, 1, 1);
  ret |= ioperm(PORT_WAIT, 1, 1);
  if (ret) {
    printf("I/O permission error!\n");
    return 1;
   }

  // Change keyboard mode to MEDIUMRAW (Linux scan code)

  ret = ioctl(STDIN_FILENO, KDSKBMODE, K_MEDIUMRAW);
  if (ret < 0) {
    printf("ioctl(KDSKBMODE) failure!\n");
    printf("Sorry, you might activate me on X-window.\n");
    printf("Please call me in a /dev/tty console!\n");
    return 1;
   }

  // Read current "termios" parameter

  ret  = tcgetattr(STDIN_FILENO, &tm);
  ret |= tcgetattr(STDIN_FILENO, &ntm);
  if (ret < 0) {
    printf("tcgetattr() failure!\n");
    return 1;
   }

  // Setup realtime key scanning

  ntm.c_iflag &= ~(ISTRIP | INLCR | ICRNL | IGNCR | IXON | IXOFF);
  ntm.c_lflag &= ~(ECHO | ICANON | ISIG);
  ntm.c_cc[ VTIME ] = 0;		// We need no wait
  ntm.c_cc[ VMIN  ] = 0;		// Blank data is acceptable
  ret = tcsetattr(STDIN_FILENO, TCSANOW, &ntm);
  if (ret < 0) {
    printf("tcsetattr() failure!\n");
    return 1;
   }

  printf("Press ESC key to abort.\n");
  while (1) {
    ret = read(STDIN_FILENO, &key, 1);	// Try to read a data
    if (ret == 0)			// There is no data...
      continue;				// Then, scan again
    up = key & 0x80;			// Extract break bit
    scan = key & 0x7F;			// Extract linux scancode
    if (scan == LINUXSCAN_ESC && up)	// ESC key was released
      break;
    else if (scan < LINUXSCAN_A || scan > LINUXSCAN_K)
      continue;				// Ignore non "ASDFGHJK" keys
    idx = scan - LINUXSCAN_A;		// Get index number
    bit = (1 << idx);			// Calculate bit position
    if (up == 0) {			// Make key
      stat |= bit;			// Turn on bit
      beep(snd_tbl[ idx ]);		// Start sound
     }
    else {				// Break key
      stat &= ~bit;			// Turn off target bit
      if (stat == 0)			// No key is pressed?
        beep(0);			// Then, turn off speaker
     }
   }

  // Restore the original "termios"

  ret = tcsetattr(STDIN_FILENO, TCSANOW, &tm);
  if (ret < 0) {
    printf("tcsetattr() failure!\n");
    ret = 1;
   }

  // Restore keyboard mode to XLATE

  ret = ioctl(STDIN_FILENO, KDSKBMODE, K_XLATE);
  if (ret < 0) {
    printf("ioctl(KDSKBMODE) failure!\n");
    ret = 1;
   }

  return 0;
 }
